#!/usr/bin/env python
"""
Copyright (C) 2017-2018 Intel Corporation.

This software and the related documents are Intel copyrighted materials,
and your use of them is governed by the express license under which they
were provided to you (License). Unless the License provides otherwise,
you may not use, modify, copy, publish, distribute, disclose or transmit
this software or the related documents without Intel's prior written permission.

This software and the related documents are provided as is, with no express or
implied warranties, other than those that are expressly stated in the License.
"""
import os
import errno
import time
import mei
import mei.mkhi
import mei.fwu
from mei.debugfs import fixed_address_soft
import sps
from common import glob

def get_mkhi_fd():
    '''Find device with MKHI client'''
    for dev_node in mei.dev_list():
        mei_fd = mei.open_dev(dev_node)
        try:
            mei.mkhi.connect(mei_fd)
        except IOError as err:
            os.close(mei_fd)
            if err.errno != errno.ENOTTY:
                continue
        else:
            return mei_fd
        time.sleep(0.1)
        mei_fd = mei.open_dev(dev_node)
        try:
            fixed_address_soft(dev_node, True)
            mei.mkhi.connect_fixed(mei_fd)
        except IOError as err:
            os.close(mei_fd)
        else:
            return mei_fd

    raise IOError(errno.ENOENT, "No device with MKHI found", "/dev/mei*")

def get_fwu_fd():
    '''Find device with FWU client'''
    for dev_node in mei.dev_list():
        mei_fd = mei.open_dev(dev_node)
        try:
            mei.fwu.connect(mei_fd)
        except IOError:
            os.close(mei_fd)
        else:
            return mei_fd

    raise IOError(errno.ENOENT, "No device with FWU found", "/dev/mei*")

def is_supported_version(fw_ver):
    '''Check if FW version is in supported list'''
    if not fw_ver[6]:
        if fw_ver[5] in [6, 7, 8, 9, 10]:
            family = glob.ME
            return False, family
        if fw_ver[5] in [12, 13]:
            family = glob.CSME
            return False, family
        if fw_ver[5] == 11:
            family = glob.CSME
            return True, family
        elif fw_ver[5] == 4:
            family = glob.TXE
            return False, family
        elif fw_ver[5] == 3:
            family = glob.TXE
            return True, family
        elif fw_ver[5] in [1, 2] and sps.get_operating_mode() == 0xF:
            family = glob.SPS
            return False, family
        elif fw_ver[5] == 0:
            family = glob.UNKNOWN
            return True, family
        else:
            family = glob.UNKNOWN
    else:
        family = glob.SPS
        if (fw_ver[5] == 4 or fw_ver[10] == 4) and \
           (fw_ver[6] in [0xA, 0xE] or fw_ver[11] in [0xA, 0xE]):
            return True, family
    return False, family

def is_txe_vulnerable(fw_ver):
    '''Check if TXE FW version is vulnerable'''
    vers = [fw_ver[5], fw_ver[4], fw_ver[8], fw_ver[7]]
    return vers[2] < 55

def is_csme_vulnerable(fw_ver):
    '''Check if CSME FW version is vulnerable'''
    vers = [fw_ver[5], fw_ver[4], fw_ver[8], fw_ver[7]]
    return vers[2] < 55

def is_sps_vulnerable(fw_ver):
    '''Check if SPS FW version is vulnerable'''
    vers1 = [fw_ver[5], fw_ver[4], fw_ver[8], fw_ver[7]]
    vers2 = [fw_ver[10], fw_ver[9], fw_ver[13], fw_ver[12]]
    return vers1[2] < 5 or vers2[2] < 5

def get_fw_ver():
    '''Get FW version from device
       Returns (fw_ver, no_drv)
    '''
    try:
        mei_fd = get_mkhi_fd()
        fw_ver = mei.mkhi.fw_ver(mei_fd)
        os.close(mei_fd)
    except IOError as the_err:
        if the_err.errno == errno.ENOENT:
            return sps.get_fw_version(), True
        mei_fd = get_fwu_fd()
        fw_ver = mei.fwu.fw_ver_as_mkhi(mei_fd)
        os.close(mei_fd)
    return fw_ver, False

def get_fw_state():
    """Calculate FW state"""
    fw_ver, no_drv = get_fw_ver()

    ret, family = is_supported_version(fw_ver)
    if not ret:
        code = glob.NOTVULNERABLE
    elif family == glob.TXE:
        if is_txe_vulnerable(fw_ver):
            code = glob.DISCOVERY_VULNERABLE
        else:
            code = glob.DISCOVERY_NOT_VULNERABLE_PATCHED
    elif family == glob.CSME:
        if is_csme_vulnerable(fw_ver):
            code = glob.DISCOVERY_VULNERABLE
        else:
            code = glob.DISCOVERY_NOT_VULNERABLE_PATCHED
    elif family == glob.SPS:
        if is_sps_vulnerable(fw_ver):
            code = glob.DISCOVERY_VULNERABLE
        else:
            code = glob.DISCOVERY_NOT_VULNERABLE_PATCHED
    else:
        code = glob.DISCOVERY_UNKNOWN

    if family == glob.SPS:
        ver_str = "%d %d.%d.%d.%d %d %d.%d.%d.%d" % (fw_ver[6], fw_ver[5],
                                                     fw_ver[4], fw_ver[8], fw_ver[7],
                                                     fw_ver[11], fw_ver[10],
                                                     fw_ver[9], fw_ver[13], fw_ver[12])
    else:
        ver_str = "%d.%d.%d.%d" % (fw_ver[5], fw_ver[4], fw_ver[8], fw_ver[7])

    return ver_str, code, family, no_drv
