#!/usr/bin/env python
"""
Copyright (C) 2018 Intel Corporation.

This software and the related documents are Intel copyrighted materials,
and your use of them is governed by the express license under which they
were provided to you (License). Unless the License provides otherwise,
you may not use, modify, copy, publish, distribute, disclose or transmit
this software or the related documents without Intel's prior written permission.

This software and the related documents are provided as is, with no express or
implied warranties, other than those that are expressly stated in the License.

Fixed HECI communication utils
"""
import sys
import time
import errno
from datetime import datetime
import struct
from sps import lpc
from sps import pci
from mei import mkhi

ME_SUBSYSTEM_BAR0_OFFSET = 0x10
ME_SUBSYSTEM_BAR1_OFFSET = 0x14

# HECI addresses and defines
H_CB_WW = 0x0 # Host Circular Buffer Write Window
H_CSR = 0x4 # Host Control Status
ME_CB_RW = 0x8 # ME Circular Buffer Read Window
ME_CSR_HA = 0xC # ME Control Status Host Access

# register bits - H_CSR
H_CBD = 0xFF000000 # Host Circular Buffer Depth
H_CBWP = 0x00FF0000 # Host Circular Buffer Write Pointer
H_CBRP = 0x0000FF00 # Host Circular Buffer Read Pointer
H_RST = 0x00000010 # Host Reset
H_RDY = 0x00000008 # Host Ready
H_IG = 0x00000004 # Host Interrupt Generate
H_IS = 0x00000002 # Host Interrupt Status
H_IE = 0x00000001 # Host Interrupt Enable

# register bits - ME_CSR_HA
ME_CBD_HRA = 0xFF000000 # ME Circular Buffer Depth Host Read Access
ME_CBWP_HRA = 0x00FF0000 # ME Circular Buffer Write Pointer Host Read Access
ME_CBRP_HRA = 0x0000FF00 # ME Circular Buffer Read Pointer Host Read Access
ME_RST_HRA = 0x00000010 # ME Reset Host Read Access
ME_RDY_HRA = 0x00000008 # ME Ready Host Read Access
ME_IG_HRA = 0x00000004 # ME Interrupt Generate Host Read Access
ME_IS_HRA = 0x00000002 # ME Interrupt Status Host Read Access
ME_IE_HRA = 0x00000001 # ME Interrupt Enable Host Read Access

HECI_CPS = 50  # minimum delay between HECI commands
MAX_TIMEOUT = 500 # ms
MAX_HECI_CMD_RETRIES = 3
SLOT_BYTE_SIZE = 4

class HeciState:
    disabled = 0
    enabled = 1
    resetting = 2

class Heci1:
    device_addr = 0
    h_csr = 0
    me_csr_ha = 0
    state = HeciState().disabled
    bdf = lpc.BDF()
    init_reset_done = False

class HeciHeader(object):
    def __init__(self, value=0):
        self.value = value
        self.me_client_id = value & 0xFF
        self.host_addr = (value & 0xFF00) >> 8
        self.length = (value & 0x1FF0000) >> 16
        self.msg_complete = value & 0x80000000 == 0x80000000

class HeciMessage(object):
    def __init__(self, header, payload):
        self.header = header
        self.payload = payload

class HeciError(Exception):
    '''Base class for HECI exceptions'''
    pass

class BufferOverflow(HeciError):
    ''' Buffer overflow detected '''
    pass

class MESideNotReady(HeciError):
    ''' Buffer overflow detected '''
    pass

class Timeout(HeciError):
    ''' Timeout expired '''
    pass

def get_time(start):
    ''' returns time span between now and start in ms '''
    delta = datetime.now() - start
    return ((delta.days * 24 * 3600 + delta.seconds) * 1000000 + delta.microseconds) // 1000.0

def get_heci1_device_address():
    ''' gets HECI-1 device BAR address '''

    bus = Heci1.bdf.bus
    dev = Heci1.bdf.dev
    funct = Heci1.bdf.funct
    heci1_bar0 = pci.read_pci_register(bus, dev, funct, ME_SUBSYSTEM_BAR0_OFFSET)
    heci_address = heci1_bar0 & 0xFFFFFFF0
    is_64bit_heci_addressing = (heci1_bar0 & 0x6) >> 1 == 0x02
    if is_64bit_heci_addressing:
        heci1_bar1 = pci.read_pci_register(bus, dev, funct, ME_SUBSYSTEM_BAR1_OFFSET)
        is_64bit_env = sys.maxsize > 2 ** 32
        if heci1_bar1 and not is_64bit_env:
            raise IOError(errno.ENOENT, "64-bit HECI-1 addressing on 32-bit environment")
        heci_address += heci1_bar1 << 32

    return heci_address

def is_ME_up(timeout=MAX_TIMEOUT):
    ''' checks if ME is up after reset '''
    start = datetime.now()
    while get_time(start) < timeout:
        Heci1.me_csr_ha = pci.mem_io_read_32(Heci1.device_addr + ME_CSR_HA)
        if not Heci1.me_csr_ha & ME_RST_HRA and Heci1.me_csr_ha & ME_RDY_HRA:
            return True
        time.sleep(0.01)

    return False

def ME_and_HOST_ready():
    ''' checks if ME and HOST are ready to communicate '''
    Heci1.me_csr_ha = pci.mem_io_read_32(Heci1.device_addr + ME_CSR_HA)
    if not Heci1.me_csr_ha & ME_RST_HRA and Heci1.me_csr_ha & ME_RDY_HRA:
        Heci1.h_csr = pci.mem_io_read_32(Heci1.device_addr + H_CSR)

    return Heci1.h_csr & H_RDY

def heci_reset():
    ''' performs HECI-1 reset '''
    Heci1.state = HeciState.resetting
    Heci1.h_csr = pci.mem_io_read_32(Heci1.device_addr + H_CSR)

    # set initiate reset sequence
    Heci1.h_csr |= H_RST
    Heci1.h_csr |= H_IG

    # Write back to H_CSR register after the bits change and read to see the result
    pci.mem_io_write_32(Heci1.device_addr + H_CSR, Heci1.h_csr)
    Heci1.h_csr = pci.mem_io_read_32(Heci1.device_addr + H_CSR)

    # make sure that H_RST is set
    if not Heci1.h_csr & H_RST:
        return False

    if not is_ME_up():
        return False

    # set proper bits
    Heci1.h_csr &= ~H_RST
    Heci1.h_csr |= H_RDY
    Heci1.h_csr |= H_IG

    pci.mem_io_write_32(Heci1.device_addr + H_CSR, Heci1.h_csr)
    Heci1.h_csr = pci.mem_io_read_32(Heci1.device_addr + H_CSR)

    # make sure H_RDY is set
    if not Heci1.h_csr & H_RDY:
        return False

    # set flag in case it's was a first reset
    Heci1.init_reset_done = True

    Heci1.state = HeciState.enabled
    
    return True

def heci_init(timeout):
    ''' initializes HECI-1 '''

    if Heci1.state != HeciState.disabled:
        return True

    start = datetime.now()

    if not Heci1.device_addr:
        Heci1.device_addr = get_heci1_device_address()

    while get_time(start) < timeout:
        Heci1.me_csr_ha = pci.mem_io_read_32(Heci1.device_addr + ME_CSR_HA)
        if not Heci1.me_csr_ha & ME_RST_HRA:
            break
        if not heci_reset():
            return False
        time.sleep(0.01)

    if get_time(start) > timeout:
        raise Timeout()

    # set up proper bit values for initialization
    Heci1.h_csr &= ~H_RST
    Heci1.h_csr |= H_RDY
    Heci1.h_csr |= H_IS
    Heci1.h_csr &= ~H_IE
    Heci1.h_csr |= H_IG

    pci.mem_io_write_32(Heci1.device_addr + H_CSR, Heci1.h_csr)
    Heci1.h_csr = pci.mem_io_read_32(Heci1.device_addr + H_CSR)

    # check if Host is ready
    if not Heci1.h_csr & H_RDY:
        return False

    Heci1.state = HeciState.enabled

    return True

def heci_uninit():
    Heci1.currentState = HeciState.disabled

def ME_CB_read_slots(length):
    ''' read slots from ME circular buffer '''

    slots_to_read = (length + 3) // 4
    output = b''
    while slots_to_read:
        output += struct.pack("<I", pci.mem_io_read_32(Heci1.device_addr + ME_CB_RW))
        slots_to_read -= 1

    # Signal to FW that SW has consumed content of ME CB
    Heci1.h_csr |= H_IG
    pci.mem_io_write_32(Heci1.device_addr + H_CSR, Heci1.h_csr)

    return output

def get_ME_CB_filled_slots():
    ''' counts ME Circular Buffer filled slots '''

    Heci1.me_csr_ha = pci.mem_io_read_32(Heci1.device_addr + ME_CSR_HA)
    read_ptr = (Heci1.me_csr_ha & ME_CBRP_HRA) >> 8
    write_ptr = (Heci1.me_csr_ha & ME_CBWP_HRA) >> 16
    buffer_depth = (Heci1.me_csr_ha & ME_CBD_HRA)  >> 24
    filled_slots = (write_ptr - read_ptr) & 0xFF

    if filled_slots > buffer_depth:
        raise BufferOverflow()

    return filled_slots

def check_HOST_CB_empty_slots(length):
    ''' checks empty slots in host circular buffer'''

    slots_needed = (length + 3) // 4
    Heci1.h_csr = pci.mem_io_read_32(Heci1.device_addr + H_CSR)

    # Calculate number of the filled and the empty slots
    read_ptr = (Heci1.h_csr & H_CBRP) >> 8
    write_ptr = (Heci1.h_csr & H_CBWP) >> 16
    buffer_depth = (Heci1.h_csr & H_CBD) >> 24
    filled_slots = (write_ptr - read_ptr) & 0xFF
    empty_slots = buffer_depth - filled_slots

    if filled_slots > buffer_depth:
        raise BufferOverflow()

    return slots_needed <= empty_slots

def send_request(msg, timeout):
    start = datetime.now()
    
    while True:
        # check wheather ME and host side are ready or if it's first request
        if not ME_and_HOST_ready() or not Heci1.init_reset_done:
            ret = heci_reset()
            if not ret:
                return False

        if get_time(start) > timeout:
            raise Timeout()

        slots_available = False
        while get_time(start) < timeout:
            # Determine number of empty slots in HECI Host Circular Buffer
            try:
                slots_available = check_HOST_CB_empty_slots(len(msg.payload))
            except BufferOverflow():
                ret = heci_reset()
                if not ret:
                    return False
            if slots_available:
                break

        # repeat procedure if not enough slots
        if not slots_available:
            continue

        # time check
        if get_time(start) > timeout:
            raise Timeout()

        # Write header
        pci.mem_io_write_32(Heci1.device_addr + H_CB_WW, msg.header.value)

        # Write payload
        length = len(msg.payload)
        offset = 0
        while length >= SLOT_BYTE_SIZE:
            buff = struct.unpack("<I", msg.payload[offset : offset + SLOT_BYTE_SIZE])[0]
            pci.mem_io_write_32(Heci1.device_addr + H_CB_WW, buff)
            offset += SLOT_BYTE_SIZE
            length -= SLOT_BYTE_SIZE
        if length > 0:
            buff = bytearray(SLOT_BYTE_SIZE - length)
            buff = struct.unpack("<I", buff.extend(msg.payload[offset:]))[0]
            pci.mem_io_write_32(Heci1.device_addr + H_CB_WW, buff)

        # Set up Host Interrupt Generate bit to inform that writing to CB is finished
        Heci1.h_csr |= H_IG
        pci.mem_io_write_32(Heci1.device_addr + H_CSR, Heci1.h_csr)
        time.sleep(0.01)
        Heci1.me_csr_ha = pci.mem_io_read_32(Heci1.device_addr + ME_CSR_HA)

        # Check ME_RDY bit
        if not Heci1.me_csr_ha & ME_RDY_HRA:
            return False

        # all semms to be fine
        return True

    return False

def get_response(me_client_id, timeout):
    slots = 0
    start = datetime.now()

    if Heci1.state == HeciState.disabled:
        return []

    while get_time(start) < timeout:
        Heci1.me_csr_ha = pci.mem_io_read_32(Heci1.device_addr + ME_CSR_HA)
        # check if ME is still in ready state or if HECI reset is needed
        if not Heci1.me_csr_ha & ME_RDY_HRA:
            ret = heci_reset()
            if not ret:
                return []

        # Check slots avalable for reading
        slots = get_ME_CB_filled_slots()
        if slots:
            break
        time.sleep(0.01)

    if not slots:
        raise Timeout()

    heci_header = None
    while slots:
        # Read message header
        if not heci_header:
            heci_header = HeciHeader(pci.mem_io_read_32(Heci1.device_addr + ME_CB_RW))
            slots -= 1

        while get_time(start) < timeout and slots * SLOT_BYTE_SIZE < heci_header.length:
            Heci1.me_csr_ha = pci.mem_io_read_32(Heci1.device_addr + ME_CSR_HA)
            # check if ME is still in ready state
            if not Heci1.me_csr_ha & ME_RDY_HRA:
                raise MESideNotReady()
            slots = get_ME_CB_filled_slots()
            time.sleep(0.01)

        # check time
        if get_time(start) > timeout:
            raise Timeout()

        if heci_header.me_client_id == me_client_id:
            data = ME_CB_read_slots(heci_header.length)
            if data:
                return data[:heci_header.length]

        # reset the number of slots and header
        get_ME_CB_filled_slots()
        heci_header = None

    return []

def send_receive_message(msg, timeout):
    ''' sends request and tries to fetch response '''
    retry_counter = 0
    curr_cmd_time = 0
    last_cmd_time = datetime.now()

    for retry_counter in range(0, MAX_HECI_CMD_RETRIES):
        # reset timer (timeout reffers to single request-response)
        start = datetime.now()

        if not heci_init(timeout):
            continue

        # don't postpone first cmd sending
        if retry_counter:
            curr_cmd_time = get_time(last_cmd_time)
            if curr_cmd_time < HECI_CPS:
                delay = HECI_CPS - curr_cmd_time
                time.sleep(delay // 1000.0)
            last_cmd_time = datetime.now()

        try:
            if send_request(msg, timeout):
                resp = get_response(msg.header.me_client_id, timeout - get_time(start))
                if resp:
                    heci_uninit()
                    return resp
        except Timeout:
            continue
        except (MESideNotReady, BufferOverflow):
            if not heci_reset():
                break
    heci_uninit()

    return []

def get_fw_version(platform):
    ''' gets FW version via MMIO heci communication '''
    heci_header = HeciHeader(0x80040007)
    mkhi_header = struct.pack("<I", 0x000002FF)
    msg = HeciMessage(heci_header, mkhi_header)
    Heci1.bdf = platform.heci_bdf
    resp = send_receive_message(msg, MAX_TIMEOUT)

    try:
        return mkhi.fw_ver_parse(resp)
    except IOError:
        return None
