#!/usr/bin/env python
"""
Copyright (C) 2018 Intel Corporation.

This software and the related documents are Intel copyrighted materials,
and your use of them is governed by the express license under which they
were provided to you (License). Unless the License provides otherwise,
you may not use, modify, copy, publish, distribute, disclose or transmit
this software or the related documents without Intel's prior written permission.

This software and the related documents are provided as is, with no express or
implied warranties, other than those that are expressly stated in the License.

PCI utils
"""

import ctypes
import os
import mmap
import sys
import struct

HOST_BRIDGE_BUS              =  0
HOST_BRIDGE_DEV              =  0
HOST_BRIDGE_FUNC             =  0
HOST_BRIDGE_PCIEXBAR_OFFSET  =  0x60

def open_pci_dev(bus, device, function, mode):
    path = "/sys/devices/pci0000:%02x/0000:%02x:%02x.%01x/config" % (bus, bus, device, function)
    fd = os.open(path, mode)
    return fd

def close_pci_dev(fd):
    os.close(fd)

def read_pci_register(bus, device, function, offset):
    try:
        fd = open_pci_dev(bus, device, function, os.O_RDWR)
        os.lseek(fd, offset, os.SEEK_SET)
        result = struct.unpack("<I", os.read(fd, 4))[0]
        close_pci_dev(fd)
    except (OSError, IOError) as the_err:
        read_pci = ctypes.CDLL('./sps/io.so').read_pci
        read_pci.restype = ctypes.c_uint32
        read_pci.argtype = ctypes.c_uint32
        addr = 0x80000000 | (bus << 16) | (device << 11) | (function << 8) | (offset & ~0x3)
        result = read_pci(addr)

    return result

def write_pci_register(bus, device, function, offset, value):
    try:
        fd = open_pci_dev(bus, device, function, os.O_WRONLY)
        os.lseek(fd, offset, os.SEEK_SET)
        os.write(fd, struct.pack("<I", value))
        close_pci_dev(fd)
    except (OSError, IOError) as the_err:
        write_pci = ctypes.CDLL('./sps/io.so').write_pci
        write_pci.restypes = ctypes.c_uint32, ctypes.c_uint32
        addr = 0x80000000 | (bus << 16) | (device << 11) | (function << 8) | (offset & ~0x3)
        data = struct.pack("<I", value)
        write_pci(data, addr)
        
def mem_io_read_32(address):
    ''' reads memory mapped pci register '''
    read = ctypes.CDLL('./sps/io.so').mem_io_read_32
    read.restype = ctypes.c_uint32
    read.argtype = ctypes.c_uint64
    addr = ctypes.c_uint64(address)
    
    return read(addr)
    
def mem_io_write_32(address, value):
    ''' writes to memory mapped pci register '''
    write = ctypes.CDLL('./sps/io.so').mem_io_write_32
    write.argtype = ctypes.c_uint64, ctypes.c_uint32
    write(ctypes.c_uint64(address), ctypes.c_uint32(value))
    
def get_pcie_mmio_addr(bus, device, function, offset):
    ''' gets physical address from PCI Express Base Address Field '''
    is_64bit = sys.maxsize > 2 ** 32
    PCIE_BAR_MASK = 0x7F
    PCIE_LENGTH_FIELD_MASK = 0x6
    pcie_bar_lower = read_pci_register(HOST_BRIDGE_BUS, HOST_BRIDGE_DEV, HOST_BRIDGE_FUNC, HOST_BRIDGE_PCIEXBAR_OFFSET)
    pcie_bar_upper = read_pci_register(HOST_BRIDGE_BUS, HOST_BRIDGE_DEV, HOST_BRIDGE_FUNC, HOST_BRIDGE_PCIEXBAR_OFFSET + 4)
    pcie_addr = 0
    if is_64bit:
        pcie_bar_upper &= PCIE_BAR_MASK
        pcie_bar_upper <<= 32
    else:
        if pcie_bar_upper & PCIE_BAR_MASK:
            #print "Unexpected value reading PCIe BAR Register (0x%08X%08X)" % (pcie_bar_upper, pcie_bar_lower)
            return 0
    pcie_bar_lower_length = (pcie_bar_lower & PCIE_LENGTH_FIELD_MASK) >> 1

    return {
        0: pcie_addr + (pcie_bar_lower & 0xF0000000) + (bus << 20) + (device << 15) + (function << 12) + offset,
        1: pcie_addr + (pcie_bar_lower & 0xF8000000) + (bus << 20) + (device << 15) + (function << 12) + offset,
        2: pcie_addr + (pcie_bar_lower & 0xFC000000) + (bus << 20) + (device << 15) + (function << 12) + offset,
        }.get(pcie_bar_lower_length, 0)

def read_register_via_pcie(bus, device, function, offset):
    ''' reads PCI register via PCIe '''
    pcie_addr = get_pcie_mmio_addr(bus, device, function, offset)
    if pcie_addr == 0:
        return 0
    return mem_io_read_32(pcie_addr)

def write_register_via_pcie(bus, device, function, offset, value):
    ''' reads PCI register via PCIe '''
    pcie_addr = get_pcie_mmio_addr(bus, device, function, offset)
    if pcie_addr == 0:
        return 0
    return mem_io_write_32(pcie_addr, value)
